"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var cache_1 = require("./cache");
// tslint:disable-next-line: no-empty
function doNothing() { }
/**
 * 删除数组中的元素
 * @param array 数组
 * @param value 值
 */
function arrayRemove(array, value) {
    var index = array.indexOf(value);
    if (index >= 0) {
        array.splice(index, 1);
    }
}
/**
 * 默认缓存索引生成函数,
 * 使用 `url`,`method`,`responseType`,`dataType`,`filePath`,`name`参数 + `data`或`formData`构建缓存Key
 * 请求的`header` 默认会被忽略
 * @param opts 请求参数对象
 * @template TOptions 完整请求参数
 * @returns string
 */
function defaultKeyBuilder(opts) {
    /**
     * 缓存的请求字段
     * https://developers.weixin.qq.com/miniprogram/dev/api/wx.request.html
     */
    var CACHE_FIELDS = [
        'url',
        'method',
        'responseType',
        'dataType',
        'filePath',
        'name' // upload
        // 'header'
        // 'data'
    ];
    var data = opts.data || opts.formData;
    return JSON.stringify(opts, CACHE_FIELDS) + (data ? JSON.stringify(data) : '');
}
exports.defaultKeyBuilder = defaultKeyBuilder;
/**
 * 是否为2xx数据
 * @param res 完整返回数据
 */
function isOkResult(res) {
    return res && res.statusCode >= 200 && res.statusCode < 300;
}
exports.isOkResult = isOkResult;
/**
 * 缓存操作,
 * 维护缓存结果,自动合并同样请求的并发操作
 * @template TRes 操作结果回调数据类型
 * @template TOptions 参数数据类型
 * @template TTask 微信任务类型
 */
var CacheOperator = /** @class */ (function () {
    /**
     * @param operator 底层操作
     * @param config 默认配置
     */
    function CacheOperator(operator, config) {
        this.cache = new cache_1.Cache();
        /**
         * 正在处理的回调
         */
        this.callbackListMap = {};
        /**
         * 处理完的回调,待删除
         */
        this.completeMap = {};
        this.op = operator;
        this.config = config || {
            expire: 15 * 60 * 1000,
            resultCondition: isOkResult
        };
    }
    /**
     * 快速创建一个
     */
    CacheOperator.createHandler = function (operator, config) {
        var cacheOperator = new CacheOperator(operator, config);
        return cacheOperator.handle.bind(cacheOperator);
    };
    /**
     * 缓存处理
     * @param options - 参数
     */
    CacheOperator.prototype.handle = function (options) {
        var _this = this;
        var keyBuilder = options.keyBuilder || this.config.keyBuilder || defaultKeyBuilder;
        var key = keyBuilder(options);
        if (!key) {
            // 不缓存
            return this.op(options);
        }
        var result = this.cache.get(key);
        if (result) {
            // 缓存命中
            result.cache = (result.cache || 0) + 1;
            try {
                if (options.success) {
                    options.success(result);
                }
            }
            catch (error) {
                this.cache.delete(key);
            }
            if (options.complete) {
                options.complete(result);
            }
        }
        else if (this.callbackListMap[key]) {
            // 请求已发送过
            var callback = this.callbackListMap[key];
            if (options.success) {
                callback.success.push(options.success);
            }
            if (options.fail) {
                callback.fail.push(options.fail);
            }
            if (options.complete) {
                callback.complete.push(options.complete);
            }
        }
        else {
            // 请求未发送过
            this.callbackListMap[key] = {
                success: options.success ? [options.success] : [],
                fail: options.fail ? [options.fail] : [],
                complete: options.complete ? [options.complete] : []
            };
            var data = __assign({}, options, { success: function (res) {
                    var expire = options.expire || _this.config.expire;
                    // 过期时间为0不缓存,但是会合并请求
                    if (expire > 0 && (options.resultCondition || _this.config.resultCondition)(res)) {
                        // 缓存请求结果
                        _this.cache.set(key, res, expire);
                    }
                    _this._getMapBeforeComplete(key).success
                        .forEach(function (v) { v(res); });
                }, fail: function (res) {
                    _this._getMapBeforeComplete(key).fail
                        .forEach(function (v) { v(res); });
                }, complete: function (res) {
                    _this.completeMap[key].forEach(function (v) { v(res); });
                    // tslint:disable-next-line: no-dynamic-delete
                    delete _this.completeMap[key];
                } });
            return this.op(data);
        }
        // tslint:disable-next-line: no-object-literal-type-assertion
        return {
            abort: function () {
                if (_this.callbackListMap[key]) {
                    if (options.success) {
                        arrayRemove(_this.callbackListMap[key].success, options.success);
                    }
                    var callbackList = [];
                    if (options.fail) {
                        arrayRemove(_this.callbackListMap[key].fail, options.fail);
                        callbackList.push(options.fail);
                    }
                    if (options.complete) {
                        arrayRemove(_this.callbackListMap[key].complete, options.complete);
                        callbackList.push(options.complete);
                    }
                    var res_1 = { errMsg: 'request:fail abort', cancel: true, source: CacheOperator.name };
                    callbackList.forEach(function (f) { f(res_1); });
                }
            },
            onHeadersReceived: doNothing,
            onProgressUpdate: doNothing
        };
    };
    /**
     * fixed #10
     * 在回调中再次发起操作前清除任务
     * @param key cacheKey
     */
    CacheOperator.prototype._getMapBeforeComplete = function (key) {
        // remove the MapList from the `callbackMapList`
        var list = this.callbackListMap[key];
        // tslint:disable-next-line: no-dynamic-delete
        delete this.callbackListMap[key];
        this.completeMap[key] = list.complete;
        return list;
    };
    return CacheOperator;
}());
exports.CacheOperator = CacheOperator;
//# sourceMappingURL=cache-operator.js.map